import ipaddress
import json
import re

from flask import Blueprint, Response

from inventory_provider.routes import common

routes = Blueprint("inventory-data-classifier-support-routes", __name__)


class ClassifierRequestError(Exception):
    status_code = 500

    def __init__(self):
        super().__init__()
        self.message = "Unclassified Internal Error"


class ClassifierProcessingError(ClassifierRequestError):
    status_code = 422

    def __init__(self, message, status_code=None):
        super().__init__()
        self.message = str(message)
        if status_code is not None:
            self.status_code = status_code


@routes.errorhandler(ClassifierRequestError)
def handle_request_error(error):
    return Response(
        response=error.message,
        status=error.status_code)


def base_interface_name(interface):
    m = re.match(r'(.*?)(\.\d+)?$', interface)
    assert m  # sanity: anything should match
    return m.group(1)


def related_interfaces(hostname, interface):
    r = common.get_redis()
    prefix = 'netconf-interfaces:%s:' % hostname
    for k in r.keys(prefix + base_interface_name(interface) + '*'):
        k = k.decode('utf-8')
        assert k.startswith(prefix)  # sanity
        assert len(k) > len(prefix)  # sanity (contains at least an interface)
        yield k[len(prefix):]


@routes.route("/trap-metadata/<source_equipment>/<path:interface>",
              methods=['GET', 'POST'])
@common.require_accepts_json
def get_trap_metadata(source_equipment, interface):
    r = common.get_redis()

    cache_key = 'classifier:cache:%s:%s' % (source_equipment, interface)
    result = r.get(cache_key)

    if result:
        result = result.decode('utf-8')
    else:
        result = {}

        services = r.get(
            'opsdb:interface_services:%s:%s' % (source_equipment, interface))
        if services:
            result['services'] = json.loads(services.decode('utf=8'))

        ifc_info = r.get(
            'netconf-interfaces:%s:%s' % (source_equipment, interface))
        if ifc_info:
            result['interface'] = json.loads(ifc_info.decode('utf-8'))

        def _related_services():
            for related in related_interfaces(source_equipment, interface):
                s = r.get('opsdb:interface_services:%s:%s'
                          % (source_equipment, related))
                if s:
                    yield json.loads(s.decode('utf-8'))

        related_services = list(_related_services())
        if related_services:
            result['related-services'] = related_services

        if not result:
            return Response(
                response="no available info for {} {}".format(
                    source_equipment, interface),
                status=404,
                mimetype="text/html")

        result = json.dumps(result)
        # cache this data for the next call
        r.set(cache_key, result.encode('utf-8'))

    return Response(result, mimetype="application/json")


def ix_peering_info(peer_info):
    """
    TODO: this is probably the least efficient way of doing this
          (if it's a problem, pre-compute these lists)

    :param peer_info: an element from ix_public_peer:address
    :return:
    """

    result = {
        'peer': peer_info,
        'group': [],
        'router': []
    }

    try:
        address = ipaddress.ip_address(peer_info['name'])
    except ValueError:
        raise ClassifierProcessingError(
            'unable to parse %r as an ip address' % address)

    description = peer_info['description']
    assert description is not None  # sanity

    protocol = type(address).__name__
    keyword = description.split(' ')[0]  # regex needed??? (e.g. tabs???)

    r = common.get_redis()

    for k in r.keys('ix_public_peer:*'):
        other = r.get(k.decode('utf-8')).decode('utf-8')
        other = json.loads(other)

        peer_address = ipaddress.ip_address(other['name'])
        if protocol != type(peer_address).__name__:
            continue

        if other['router'] == peer_info['router']:
            result['router'].append(other['name'])

        assert other['description'] is not None  # sanity: as above...
        if other['description'].startswith(keyword):
            result['group'].append(other['name'])

    return result


def find_interfaces(address):
    """
    TODO: this is probably the least efficient way of doing this
          (if it's a problem, pre-compute these lists)

    :param address: an ipaddress object
    :return:
    """
    r = common.get_redis()
    for k in r.keys('reverse_interface_addresses:*'):
        info = r.get(k.decode('utf-8')).decode('utf-8')
        info = json.loads(info)
        interface = ipaddress.ip_interface(info['interface address'])
        if address in interface.network:
            yield info


def find_interfaces_and_services(address_str):
    """

    :param address_str: an ipaddress object
    :return:
    """

    try:
        address = ipaddress.ip_address(address_str)
    except ValueError:
        raise ClassifierProcessingError(
            'unable to parse %r as an ip address' % address_str)

    r = common.get_redis()
    for interface in find_interfaces(address):

        services = r.get(
            'opsdb:interface_services:%s:%s' % (
                interface['router'],
                interface['interface name']))
        if not services:
            services = []
        else:
            services = json.loads(services.decode('utf=8'))

        yield {
            'interface': interface,
            'services': services
        }


@routes.route("/peer-info/<address>", methods=['GET', 'POST'])
@common.require_accepts_json
def peer_info(address):

    r = common.get_redis()

    cache_key = 'classifier:peer-cache:%s' % address

    result = r.get(cache_key)
    if result:
        result = result.decode('utf-8')
    else:

        result = {}

        info = r.get('ix_public_peer:%s' % address)
        if info:
            info = info.decode('utf-8')
            result['ix-public-peer-info'] = ix_peering_info(json.loads(info))

        info = r.get('vpn_rr_peer:%s' % address)
        if info:
            info = info.decode('utf-8')
            result['vpn-rr-peer-info'] = json.loads(info)

        interfaces = list(find_interfaces_and_services(address))
        if interfaces:
            result['interfaces'] = interfaces

        if not result:
            return Response(
                response='no peering info found for %s' % address,
                status=404,
                mimetype="text/html")

        result = json.dumps(result)
        # cache this data for the next call
        r.set(cache_key, result.encode('utf-8'))

    return Response(result, mimetype="application/json")
